@@ -12,14 +12,42 @@ require_relative 'catalog-util/fileresources'
 require_relative 'errors'
 
 module OctocatalogDiff
-  # This class represents a catalog. Generation of the catalog is handled via one of the
+  # Basic methods for interacting with a catalog. Generation of the catalog is handled via one of the
   # supported backends listed above as 'require_relative'. Usually, the 'computed' backend
   # will build the catalog from the Puppet command.
   class Catalog
-    # Readable
-    attr_reader :built, :catalog, :catalog_json
+    attr_accessor :node
+    attr_reader :built, :catalog, :catalog_json, :options
 
     # Constructor
+    def initialize(options = {})
+      unless options.is_a?(Hash)
+        raise ArgumentError, "#{self.class}.initialize requires hash argument, not #{options.class}"
+      end
+      @options = options
+
+      # Basic settings
+      @node = options[:node]
+      @error_message = nil
+      @catalog = nil
+      @catalog_json = nil
+      @retries = nil
+
+      # The compilation directory can be overridden, e.g. when testing
+      @override_compilation_dir = options[:compilation_dir]
+
+      # Keep track of whether references have been validated yet. Allow this to be fudged for when we do
+      # not desire reference validation to happen (e.g., for the "from" catalog that is otherwise valid).
+      @references_validated = options[:references_validated] || false
+
+      # Keep track of whether file resources have been converted.
+      @file_resources_converted = false
+
+      # Keep track of whether it's built yet
+      @built = false
+    end
+
+    # Guess the backend from the input and return the appropriate catalog object.
     # @param :backend [Symbol] If set, this will force a backend
     # @param :json [String] JSON catalog content (will avoid running Puppet to compile catalog)
     # @param :puppetdb [Object] If set, pull the catalog from PuppetDB rather than building
@@ -30,18 +58,25 @@ module OctocatalogDiff
     # @param :pass_env_vars [Array<String>] OPTIONAL: Additional environment vars to pass
     # @param :convert_file_resources [Boolean] OPTIONAL: Convert file resource source to content
     # @param :storeconfigs [Boolean] OPTIONAL: Pass the '-s' flag, for puppetdb (storeconfigs) integration
-    def initialize(options = {})
-      @options = options
+    # @return [OctocatalogDiff::Catalog::<?>] Catalog object from guessed backend
+    def self.create(options = {})
+      # Hard-coded backend
+      if options[:backend]
+        return OctocatalogDiff::Catalog::JSON.new(options) if options[:backend] == :json
+        return OctocatalogDiff::Catalog::PuppetDB.new(options) if options[:backend] == :puppetdb
+        return OctocatalogDiff::Catalog::PuppetMaster.new(options) if options[:backend] == :puppetmaster
+        return OctocatalogDiff::Catalog::Computed.new(options) if options[:backend] == :computed
+        return OctocatalogDiff::Catalog::Noop.new(options) if options[:backend] == :noop
+        raise ArgumentError, "Unknown backend :#{options[:backend]}"
+      end
 
-      # Call appropriate backend for catalog generation
-      @catalog_obj = backend(options)
+      # Determine backend based on arguments
+      return OctocatalogDiff::Catalog::JSON.new(options) if options[:json]
+      return OctocatalogDiff::Catalog::PuppetDB.new(options) if options[:puppetdb]
+      return OctocatalogDiff::Catalog::PuppetMaster.new(options) if options[:puppet_master]
 
-      # The catalog is not built yet, except if the backend has no build method
-      @built = false
-      build unless @catalog_obj.respond_to?(:build)
-
-      # The compilation directory can be overridden, e.g. when testing
-      @override_compilation_dir = nil
+      # Default is to build catalog ourselves
+      OctocatalogDiff::Catalog::Computed.new(options)
     end
 
     # Build catalog - this method needs to be called to build the catalog. It is separate due to
@@ -49,43 +84,42 @@ module OctocatalogDiff
     # object so it cannot be part of any object that is passed around.
     # @param logger [Logger] Logger object, initialized to a default throwaway value
     def build(logger = Logger.new(StringIO.new))
-      # Only build once
+      # If already built, don't build again
       return if @built
       @built = true
 
-      # Call catalog's build method.
-      if @catalog_obj.respond_to?(:build)
-        logger.debug "Calling build for object #{@catalog_obj.class}"
-        @catalog_obj.build(logger)
-      end
-
-      # These methods must exist in all backends
-      @catalog = @catalog_obj.catalog
-      @catalog_json = @catalog_obj.catalog_json
-      @error_message = @catalog_obj.error_message
-
       # The resource hash is computed the first time it's needed. For now initialize it as nil.
       @resource_hash = nil
 
+      # Invoke the backend's build method, if there is one. There's a stub below in case there's not.
+      logger.debug "Calling build for object #{self.class}"
+      build_catalog(logger)
+
       # Perform post-generation processing of the catalog
       return unless valid?
-      unless @catalog_obj.respond_to?(:convert_file_resources) && @catalog_obj.convert_file_resources == false
-        if @options.fetch(:compare_file_text, false)
-          OctocatalogDiff::CatalogUtil::FileResources.convert_file_resources(self, environment)
-        end
-      end
+
+      validate_references
+      return unless valid?
+
+      convert_file_resources if @options[:compare_file_text]
+
+      true
+    end
+
+    # Stub method if the backend does not contain a build method.
+    def build_catalog(_logger)
     end
 
     # Compilation environment
     # @return [String] Compilation environment (if set), else 'production' by default
     def environment
-      @catalog_obj.respond_to?(:environment) ? @catalog_obj.environment : 'production'
+      @environment ||= 'production'
     end
 
     # For logging we may wish to know the backend being used
     # @return [String] Class of backend used
     def builder
-      @catalog_obj.class.to_s
+      self.class.to_s
     end
 
     # Set the catalog JSON
@@ -98,8 +132,7 @@ module OctocatalogDiff
     # This retrieves the compilation directory from the catalog, or otherwise the passed-in directory.
     # @return [String] Compilation directory
     def compilation_dir
-      return @override_compilation_dir if @override_compilation_dir
-      @catalog_obj.respond_to?(:compilation_dir) ? @catalog_obj.compilation_dir : @options[:basedir]
+      @override_compilation_dir || @options[:basedir]
     end
 
     # The compilation directory can be overridden, e.g. during testing.
@@ -108,16 +141,16 @@ module OctocatalogDiff
       @override_compilation_dir = dir
     end
 
-    # Determine whether the underlying catalog object supports :compare_file_text
-    # @return [Boolean] Whether underlying catalog object supports :compare_file_text
-    def convert_file_resources
-      return true unless @catalog_obj.respond_to?(:convert_file_resources)
-      @catalog_obj.convert_file_resources
+    # Stub method for "convert_file_resources" -- returns false because if the underlying class does
+    # not implement this method, it's not supported.
+    def convert_file_resources(_dry_run = false)
+      false
     end
 
     # Retrieve the error message.
     # @return [String] Error message (maximum 20,000 characters) - nil if no error.
     def error_message
+      build
       return nil if @error_message.nil? || !@error_message.is_a?(String)
       @error_message[0, 20_000]
     end
@@ -133,12 +166,11 @@ module OctocatalogDiff
       @resource_hash = nil
     end
 
-    # This retrieves the version of Puppet used to compile a catalog. If the underlying catalog was not
-    # compiled by running Puppet (e.g., it was read in from JSON or puppetdb), then this returns the
-    # puppet version optionally passed in to the constructor. This can also be nil.
+    # Stub method to return the puppet version if the back end doesn't support this.
     # @return [String] Puppet version
     def puppet_version
-      @catalog_obj.respond_to?(:puppet_version) ? @catalog_obj.puppet_version : @options[:puppet_version]
+      build
+      @options[:puppet_version]
     end
 
     # This allows retrieving a resource by type and title. This is intended for use when a O(1) lookup is required.
@@ -147,6 +179,7 @@ module OctocatalogDiff
     # @return [Hash] Resource item
     def resource(opts = {})
       raise ArgumentError, ':type and :title are required' unless opts[:type] && opts[:title]
+      build
       build_resource_hash if @resource_hash.nil?
       return nil unless @resource_hash[opts[:type]].is_a?(Hash)
       @resource_hash[opts[:type]][opts[:title]]
@@ -155,6 +188,7 @@ module OctocatalogDiff
     # This is a compatibility layer for the resources, which are in a different place in Puppet 3.x and Puppet 4.x
     # @return [Array] Resource array
     def resources
+      build
       raise OctocatalogDiff::Errors::CatalogError, 'Catalog does not appear to have been built' if !valid? && error_message.nil?
       raise OctocatalogDiff::Errors::CatalogError, error_message unless valid?
       return @catalog['data']['resources'] if @catalog['data'].is_a?(Hash) && @catalog['data']['resources'].is_a?(Array)
@@ -165,16 +199,17 @@ module OctocatalogDiff
       # :nocov:
     end
 
-    # This retrieves the number of retries necessary to compile the catalog. If the underlying catalog
+    # Stub method of the the number of retries necessary to compile the catalog. If the underlying catalog
     # generation backend does not support retries, nil is returned.
     # @return [Integer] Retry count
     def retries
-      @retries = @catalog_obj.respond_to?(:retries) ? @catalog_obj.retries : nil
+      nil
     end
 
     # Determine if the catalog build was successful.
     # @return [Boolean] Whether the catalog is valid
     def valid?
+      build
       !@catalog.nil?
     end
 
@@ -182,11 +217,19 @@ module OctocatalogDiff
     # Raise a OctocatalogDiff::Errors::ReferenceValidationError for any found to be missing.
     # Uses @options[:validate_references] to influence which references are checked.
     def validate_references
+      # If we've already done the validation, don't do it again
+      return if @references_validated
+      @references_validated = true
+
       # Skip out early if no reference validation has been requested.
       unless @options[:validate_references].is_a?(Array) && @options[:validate_references].any?
         return
       end
 
+      # Puppet 5 has reference validation built-in and enabled, so there won't even be a valid catalog if
+      # there were invalid references. It's pointless to perform validation of our own.
+      return if puppet_version && puppet_version >= '5.0.0'
+
       # Iterate over all the resources and check each one that has one of the attributes being checked.
       # Keep track of all references that are missing for ultimate inclusion in the error message.
       missing = []
@@ -204,7 +247,7 @@ module OctocatalogDiff
       # At this point there is at least one broken/missing reference. Format an error message and raise.
       errors = format_missing_references(missing)
       plural = errors =~ /;/ ? 's' : ''
-      raise OctocatalogDiff::Errors::ReferenceValidationError, "Catalog has broken reference#{plural}: #{errors}"
+      self.error_message = "Catalog has broken reference#{plural}: #{errors}"
     end
 
     private
@@ -261,31 +304,21 @@ module OctocatalogDiff
         unless res =~ /\A([\w:]+)\[(.+)\]\z/
           raise ArgumentError, "Resource #{res} is not in the expected format"
         end
-        resource(type: Regexp.last_match(1), title: Regexp.last_match(2)).nil?
+
+        type = Regexp.last_match(1)
+        title = normalized_title(Regexp.last_match(2), type)
+        resource(type: type, title: title).nil?
       end
     end
 
-    # Private method: Choose backend based on passed-in options
-    # @param options [Hash] Options passed into constructor
-    # @return [Object] OctocatalogDiff::Catalog::<whatever> object
-    def backend(options)
-      # Hard-coded backend
-      if options[:backend]
-        return OctocatalogDiff::Catalog::JSON.new(options) if options[:backend] == :json
-        return OctocatalogDiff::Catalog::PuppetDB.new(options) if options[:backend] == :puppetdb
-        return OctocatalogDiff::Catalog::PuppetMaster.new(options) if options[:backend] == :puppetmaster
-        return OctocatalogDiff::Catalog::Computed.new(options) if options[:backend] == :computed
-        return OctocatalogDiff::Catalog::Noop.new(options) if options[:backend] == :noop
-        raise ArgumentError, "Unknown backend :#{options[:backend]}"
-      end
+    # Private method: Given a title string, normalize it according to the rules
+    # used by puppet 4.10.x for file resource title normalization:
+    # https://github.com/puppetlabs/puppet/blob/4.10.x/lib/puppet/type/file.rb#L42
+    def normalized_title(title_string, type)
+      return title_string if type != 'File'
 
-      # Determine backend based on arguments
-      return OctocatalogDiff::Catalog::JSON.new(options) if options[:json]
-      return OctocatalogDiff::Catalog::PuppetDB.new(options) if options[:puppetdb]
-      return OctocatalogDiff::Catalog::PuppetMaster.new(options) if options[:puppet_master]
-
-      # Default is to build catalog ourselves
-      OctocatalogDiff::Catalog::Computed.new(options)
+      matches = title_string.match(%r{^(?<normalized_path>/|.+:/|.*[^/])/*\Z}m)
+      matches[:normalized_path] || title_string
     end
 
     # Private method: Build the resource hash to be used used for O(1) lookups by type and title.
@@ -294,7 +327,9 @@ module OctocatalogDiff
       @resource_hash = {}
       resources.each do |resource|
         @resource_hash[resource['type']] ||= {}
-        @resource_hash[resource['type']][resource['title']] = resource
+
+        title = normalized_title(resource['title'], resource['type'])
+        @resource_hash[resource['type']][title] = resource
 
         if resource.key?('parameters') && resource['parameters'].key?('alias')
           @resource_hash[resource['type']][resource['parameters']['alias']] = resource
